/***************************************************************************************\
 * File:            HardwareProfile.h
 *
 * Purpose:         Hardware Platform definitions
 *
 * Target H/W:      Olimex PIC32-MX460 Development Board (plus REMI add-ons)
 *                  Olimex PIC32-MX340 Prototyping Board ( ..   ..   ..    )
 *
 * Processor:       PIC32MX460F512L or PIC32MX340F512H
 *
 * Compiler:        Microchip XC32 v1.33++ under MPLAB.X v2.20++
 *
\***************************************************************************************/

#ifndef __HARDWARE_PROFILE_H
#define __HARDWARE_PROFILE_H

#include "Common/Compiler.h"


//========================  Select Target Platform ======================================
//     <!> Processor type must also be selected in MPLAB'X Project Properties <!>
//
#ifdef BUILD_REMI_CONTROLLER_APP  //-------------------------------------------

// Comment out the following line if using a PIC32MX340 controller board...
#define USE_OLIMEX_PIC32MX460_BOARD

#else  // All other projects are using PIC32MX460 board with RTCC
#define USE_OLIMEX_PIC32MX460_BOARD
#define RTCC_TYPE_MCP79410   1

#endif  //---------------------------------------------------------------------


#define DEVICE_ID_PIC32MX460F512L  0x00978053
#define DEVICE_ID_PIC32MX340F512H  0x00916053
#define DEVICE_ID_MASK             0x0FFFFFFF
#define GET_DEVICE_ID()            (DEVID)     // Register in MCU flash PM

// Check we have the correct processor type selected in Project Properties...
#ifdef USE_OLIMEX_PIC32MX460_BOARD
#ifndef __32MX460F512L__
#error "Processor type PIC32MX460F512L must be selected!"
#endif
#else  // assume target is PIC32MX340
#ifndef __32MX340F512H__
#error "Processor type PIC32MX340F512H must be selected!"
#endif
#endif

#define CHECK_SOFTWARE_TIMER_ACCURACY  1

// kernel RTC support (software RTC) is needed if a hardware RTCC device is used.
#if defined RTCC_TYPE_ISL1208 || defined RTCC_TYPE_MCP79410
#define INCLUDE_KERNEL_RTC_SUPPORT    // (firmware build option)
#endif

#if defined RTCC_TYPE_ISL1208 && defined RTCC_TYPE_MCP79410
#error "System can define only one RTCC device type!"
#endif

//========================  Select Display Panel Type ===================================
//
#define LCD_CONTROLLER_ST7920    // Using 128x64 mono graphics LCD with ST7920 chip


//=================  Select Mass Storage Device and MDD File System  ====================
//
// Don't confuse this with the REMI built-in synth -- it's a different project!
#if defined BUILD_SOUND_FX_SYNTH_APP
#define USE_USB_MSD_HOST_INTERFACE        // using USB flash drive
#define USE_MDD_FILE_SYSTEM               // Memory Disk Drive file system (FAT32)
#define USE_SST25VF016                    // using SPI serial data flash (SDF) memory
#define USE_SDF_FILE_SYSTEM               // using SDF file system
#endif


//========================== Select other custom hardware ===============================
//
// Select the I2C peripheral channel(s) to be used in the project...
#define USE_I2C_CHANNEL_1
#define USE_I2C_CHANNEL_2
/*
#define USE_I2C_CHANNEL_3
#define USE_I2C_CHANNEL_4
*/

// Select the PIC32 I2C channel that the 24LC08 EEPROM uses...
#if defined USE_OLIMEX_PIC32MX340_BOARD
#define EEPROM_I2C_PORT_NUMBER   1    // value is 1 for I2C1
#else  // assume USE_OLIMEX_PIC32MX460_BOARD
#define EEPROM_I2C_PORT_NUMBER   2    // value is 2 for I2C2
#endif

#if defined BUILD_REMI_CONTROLLER_APP || defined BUILD_LCD_GRAPHICS_DEMO
#define USE_4_WIRE_KEYPAD
#endif

// DEBUG LED driver macros
#if defined BUILD_REMI_CONTROLLER_APP
#define DEBUG_LED_INIT()    TRISFbits.TRISF1 = 0
#define DEBUG_LED_ON()      LATFbits.LATF1 = 1
#define DEBUG_LED_OFF()     LATFbits.LATF1 = 0
#define DEBUG_LED_TOGGLE()  LATFbits.LATF1 ^= 1
#elif defined USE_OLIMEX_PIC32MX460_BOARD
#define DEBUG_LED_INIT()    TRISDbits.TRISD1 = 0
#define DEBUG_LED_ON()      LATDbits.LATD1 = 1
#define DEBUG_LED_OFF()     LATDbits.LATD1 = 0
#define DEBUG_LED_TOGGLE()  LATDbits.LATD1 ^= 1
#endif

/*********************************************************************
* SST25VF016 SPI Flash Memory configuration:
* Define here the SPI channel and slave-select (CS#) pin to use.
*
* Olimex PIC32-MX460 board has an SST25VF016B added in place of the
* RF module, interfaced to SPI channel 1. Flash CS# is RB2.
*
* SPI clk rate, Fclk = FPB / (2 x (SPIxBRG + 1))
* Example:
*     FPB = FSYS = 80MHz, SPIxBRG = 1 --> Fclk = 80 / 4 = 20 MHz
*
*********************************************************************/

#if defined (USE_SST25VF016)

#define SST25_SPI_CHANNEL     1   // SPI channel for SST25VF016
#define SPI_CHANNEL_1_ENABLE
#define SST25_SPI_SPIBRG      1   // Value put in SPIxBRG reg.

#define SST25_CS_TRIS     TRISBbits.TRISB2  // SPI Flash CS# = RB2
#define SST25_CS_LAT      LATBbits.LATB2

#endif // USE_SST25VF016


/*********************************************************************
* GetSystemClock() returns system clock frequency.
*
* GetPeripheralClock() returns peripheral clock frequency.
*
* GetInstructionClock() returns instruction clock frequency.
*
*********************************************************************/

/*********************************************************************
* Macro: #define	GetSystemClock() 
*
* Overview: This macro returns the system clock frequency in Hertz.
*
********************************************************************/

#define GetSystemClock()    (80000000ul)  
 

/*********************************************************************
* Macro: #define	GetPeripheralClock() 
*
* Overview: This macro returns the peripheral clock frequency in Hertz.
*
********************************************************************/
#define GetPeripheralClock()    (GetSystemClock() / (1 << OSCCONbits.PBDIV))


/*********************************************************************
* Macro: #define	GetInstructionClock() 
*
* Overview: This macro returns the instruction clock rate in Hertz,
*           dependent on the number of flash PM wait states configured.
*           (PIC32MX3xx/4xx requires 2 wait states @ SYSCLK = 80MHz.)
*           The formula can be tweaked using a software timing function
*           e.g. Delay10us() in Common/TimeDelay.c.
*
********************************************************************/
#define GetInstructionClock()   ((80000000 * 10) / 23)


//=======================================================================================

// Debug functions used for Microchip Library Software (MLA)
#ifdef USE_DEBUG_TRACE
////
#define DEBUG_Initialize()     // Assume debug UART iniz'd by app
#define DEBUG_PutChar(c)       putch(c)
#define DEBUG_PutString(s)     putstr(s)
#define DEBUG_PutHexUINT8(b)   putHexByte(b)
#define DEBUG_PutHexUINT16(w)  putHexWord(w)
#define DEBUG_PutHexUINT32(l)  putHexLong(l)
#define UART2PutChar(c)        putch(c)
#define UART2PrintString(s)    putstr(s)
#define UART2PutHex(b)         putHexByte(b)
#define UART2PutHexDWord(l)    putHexLong(l)
#define UART2PutDec(n)         putDecimal(n,1)
////
#else  // Debug functions not needed...
////
#define DEBUG_Initialize()
#define DEBUG_PutChar(c)
#define DEBUG_PutString(s)
#define DEBUG_PutHexUINT8(b)
#define DEBUG_PutHexUINT16(w)
#define DEBUG_PutHexUINT32(l)
#define UART2PutChar(c)
#define UART2PrintString(s)
#define UART2PutHex(b)
#define UART2PutHexDWord(l)
#define UART2PutDec(n)
////
#endif

#endif // __HARDWARE_PROFILE_H
