/**
 *   File: RTC_support.c
 *
 *   Module contains low-level support functions for real-time clock (RTC),
 *   including hardware RTCC devices.
 *
 *   RTCC chips supported:  Intersil ISL1208 (I2C), Microchip MCP79410 (I2C)
 *
 *   Author:     M.J.Bauer, 2014   [www.mjbauer.biz]
 *
 *   Identify the RTCC device installed, if any, by defining one of the symbols
 *   RTCC_TYPE_ISL1208 or RTCC_TYPE_MCP79410 in file "HardwareProfile.h".
 * 
 */
#include <stdlib.h>
#include <string.h>

#include "system_def.h"

#if defined RTCC_TYPE_ISL1208 || defined RTCC_TYPE_MCP79410
// If the above condition is true, compile down to end of file --------------------------

#include "RTC_support.h"

#ifdef  USE_CONSOLE_CLI
#include "console.h"
#endif

#ifdef RTCC_TYPE_ISL1208
#define SCRATCH_REG_ADDR  0x12
#define CONTROL_REG_ADDR  0x08
#define STATUS_REG_ADDR   0x07
#else
#define SCRATCH_REG_ADDR  0x20
#define CONTROL_REG_ADDR  0x07
#define STATUS_REG_ADDR   0x03
#endif


// System real-time clock/calendar (RTC) data...
struct  tDateTime  gsRTCbuf;   // System RTC "register" buffer for access by application
struct  tDateTime  sRTC;       // System RTC "live" date-time "registers"

static  uint8   LUT_DaysinMonth[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

char *asDayOfWeekName[] = { "SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT" };

char *asMonthName[] = { "", "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
                            "JUL", "AUG", "SEP", "OCT", "NOV", "DEC" };

short  loop;

/**
*  Background task routine performed every day at midnight.
*/
void RTC_NewDayRolloverTask(void)
{
    BYTE  yLeapYear;
    SHORT days_in_month;

    sRTC.dow++ ;
    if ( sRTC.dow >= 7 )  sRTC.dow = 0;     // it must be Sunday

    yLeapYear = ( sRTC.year % 4 == 0 && sRTC.year != 0 );
    if ( yLeapYear && sRTC.month == 2 )
        days_in_month = 29;
    else
        days_in_month = LUT_DaysinMonth[sRTC.month];

    if ( ++sRTC.day > days_in_month )
    {
        sRTC.day = 1;
        if ( ++sRTC.month > 12 )
        {
            sRTC.month = 1;
            if ( ++sRTC.year > 99 )     // year is 2099 !
                sRTC.year = 0;
        }
    }
}

/**
*  Set the real-time clock registers from date/time values stored in the global
*  RTC buffer: gsRTCbuf, which must be setup before the call.
*  The process is repeated if the "live" RTC 'seconds' register rolls over
*  during the process, to ensure that the live data are not currupted by a timer
*  tick interrupt. One re-try should suffice, but the function will loop until
*  doomsday (i.e. until watchdog reset occurs) if necessary.
*
*  The day-of-the-week value is computed from the set date and time, based on 01/01/2000
*  being a Saturday (dow = 6) and the number of days passed since that day.
*/
void  RTC_date_time_set( void )
{
    BYTE    bYear;
    SHORT   wDaysPassed = 0;  // Days passed since Saturday 01/01/2000

    do
    {
        sRTC.secs  = gsRTCbuf.secs;
        sRTC.mins  = gsRTCbuf.mins;
        sRTC.hour  = gsRTCbuf.hour;
        sRTC.day   = gsRTCbuf.day;
        sRTC.month = gsRTCbuf.month;
        sRTC.year  = gsRTCbuf.year;
    }
    while ( sRTC.secs != gsRTCbuf.secs );

    for ( bYear=0;  bYear < gsRTCbuf.year;  bYear++ )
    {
        wDaysPassed += days_in_whole_year( bYear );
    }
    wDaysPassed += day_of_this_year() - 1;
    sRTC.dow = (wDaysPassed + 6) % 7 ;        //**  DOW: 0==SUN, 1==MON, ... 6==SAT  **
}


/*
*  Read the "live" real-time clock registers into the global RTC buffer: gsRTCbuf.
*  The process is repeated if the "live" RTC seconds register rolls over
*  during the process, to ensure that the buffer's data are not currupted
*  by a timer tick interrupt.  One re-try should suffice.
*  The buffered RTC date/time is checked for validity;  if invalid,
*  the function returns FALSE, otherwise TRUE.
*/
BYTE  RTC_date_time_read( void )
{
    do
    {
        gsRTCbuf.secs  = sRTC.secs;
        gsRTCbuf.mins  = sRTC.mins;
        gsRTCbuf.hour  = sRTC.hour;
        gsRTCbuf.dow   = sRTC.dow;
        gsRTCbuf.day   = sRTC.day;
        gsRTCbuf.month = sRTC.month;
        gsRTCbuf.year  = sRTC.year;
    }
    while ( sRTC.secs != gsRTCbuf.secs );

    if ( ( gsRTCbuf.month > 12 )
    ||   ( gsRTCbuf.month == 0 )
    ||   ( gsRTCbuf.day   > 31 )
    ||   ( gsRTCbuf.day   == 0 )
    ||   ( gsRTCbuf.hour  > 23 )
    ||   ( gsRTCbuf.mins   > 59 )
    ||   ( gsRTCbuf.secs   > 59 ) )
    {
//      SET_BIT( g_SystemErrors, RTC_INVALID );
        return FALSE;
    }
    else  return TRUE;
}


/*****
*    Compute number of days in whole of year specified.
*
*    Arg(s):   (uint8) bYear = year in the range 0..99 (0 => 2000)
*    Returns:  366 if bYear is a leap year, else 365.
*/
SHORT  days_in_whole_year( uint8 bYear )
{
    if ( bYear % 4 == 0 )  return 366;    // It's a leap year
    else  return 365;
}


/*****
*    Compute the 'day-of-year" number (1..366) for the date stored in the global date.time
*    structure, gsRTCbuf.
*
*    Returns:  Day-of-year, in the range 1..365, (1..366 if it's a leap year).
*/
SHORT  day_of_this_year( void )
{
    uint8   bMonth;
    uint16  wResult = gsRTCbuf.day;
    bool    yLeapYear = ( gsRTCbuf.year % 4 == 0 );

    for ( bMonth = 1;  bMonth < gsRTCbuf.month;  bMonth++ )
    {
        if ( yLeapYear && bMonth == 2 )  wResult += 29;
        else  wResult += LUT_DaysinMonth[bMonth];
    }
    return  wResult;
}


/**
*    Function to update the MMD File System time-stamp variables from RTC time and date reg's,
*    where a "user-defined" (custom) real-time clock module is implemented.
*    Should be called immediately before FSopen() and FSclose() operations.
*    If the RTC is not set properly, the time & date default to 00:00:00 2000-01-01.
*
*    The macro USERDEFINEDCLOCK is defined in FSconfig.h.
*    The function SetClockVars() is implemented in FSIO.c.
*/
int  UpdateFSrealTimeVars(void)
{
#if defined USE_MDD_FILE_SYSTEM && defined USERDEFINEDCLOCK

    unsigned int year = 2000 + gsRTCbuf.year;
    int result = 0;

    if (RTC_date_time_read() == 0) result = ERROR;
    else result = SetClockVars(year, gsRTCbuf.month, gsRTCbuf.day,
                      gsRTCbuf.hour, gsRTCbuf.mins, gsRTCbuf.secs);

    return result;

#endif
}


/**
 * ============  HARDWARE REAL-TIME CLOCK FUNCTIONS  =========================================
 *
 * The system RTC is synchronized to a hardware battery-backed RTC chip (ISL1208 or MCP97410)
 * which interfaces to the processor via the I2C bus (I2C2).
 *
 * Console CLI commands are provided to manually set the hardware RTCC date/time and to
 * synchronize the system RTC to the hardware RTCC following a power failure.
 */

#ifdef RTCC_TYPE_ISL1208
/*
 * Initialize the hardware RTCC chip and set time & date from global RTC buffer
 * The function checks that the hardware RTCC chip is communicating properly.
 * Return value is zero (0) if initialization was successful; otherwise it returns...
 *   (-1) if an I2C bus error was detected;
 *   (-2) if RTCC SRAM register write failed to verify.
 *   (-3) if the powerfail flag (RTCF bit) is not cleared
 */
int  RTCC_Initialize( void )
{
    if (RTC_RegWrite(0x12, 0xA5) == ERROR) return (-1);  // Write SRAM reg
    if (RTC_RegRead(0x12) != 0xA5) return (-2);          // Read back & verify

    RTC_RegWrite(0x0A, 0x00);            // ATR bits = 00
    RTC_RegWrite(0x0B, 0x00);            // DTR bits = 00
    RTC_RegWrite(0x07, 0x10);            // Set WRTC bit to write date/time reg's

    RTC_RegWrite(0x00, Byte_to_BCD(gsRTCbuf.secs));
    RTC_RegWrite(0x01, Byte_to_BCD(gsRTCbuf.mins));
    RTC_RegWrite(0x02, Byte_to_BCD(gsRTCbuf.hour));
    RTC_RegWrite(0x03, Byte_to_BCD(gsRTCbuf.day));
    RTC_RegWrite(0x04, Byte_to_BCD(gsRTCbuf.month));
    RTC_RegWrite(0x05, Byte_to_BCD(gsRTCbuf.year));

    DelayMs(200);   // Allow generous delay for osc. to start up

    if (RTC_RegRead(0x07) & 0x01) return (-3);  // Check 'RTCF' bit -- should be 0

    RTC_RegWrite(0x08, 0x03);            // Enable Fout = 1024Hz
    RTC_RegWrite(0x07, 0x00);            // Clear WRTC

    return 0;
}

/*
 * Synchronize system RTC time & date with hardware RTCC
 */
int  RTCC_Synchronize( void )
{
    gsRTCbuf.secs = BCD_to_binary(RTC_RegRead(0x00) & 0x7F);
    gsRTCbuf.mins = BCD_to_binary(RTC_RegRead(0x01) & 0x7F);
    gsRTCbuf.hour = BCD_to_binary(RTC_RegRead(0x02) & 0x3F);
    gsRTCbuf.day = BCD_to_binary(RTC_RegRead(0x03)) & 0x3F;
    gsRTCbuf.month = BCD_to_binary(RTC_RegRead(0x04) & 0x1F);
    gsRTCbuf.year = BCD_to_binary(RTC_RegRead(0x05));

    RTC_date_time_set();
}

#endif // RTCC_TYPE_ISL1208


#ifdef RTCC_TYPE_MCP79410
/*
 * Initialize the hardware RTCC chip and set time & date from global RTC buffer.
 * The seconds count is zeroed.
 * The function checks that the hardware RTCC chip is communicating properly.
 * Return value is zero (0) if initialization was successful; otherwise it returns...
 *   (-1) if an I2C bus error was detected;
 *   (-2) if RTCC SRAM register write failed to verify.
 */
int  RTCC_Initialize( void )
{
    if (RTC_RegWrite(0x20, 0xA5) == ERROR) return (-1);  // Write SRAM reg
    if (RTC_RegRead(0x20) != 0xA5) return (-2);          // Read back & verify

    RTC_RegWrite(0x00, 0x00);    // Stop the clock (ST = 0)
    RTC_RegWrite(0x03, 0x08);    // VBATEN = 1
    RTC_RegWrite(0x07, 0x41);    // Enable OUT = 4.096kHz

    RTC_RegWrite(0x06, Byte_to_BCD(gsRTCbuf.year));
    RTC_RegWrite(0x05, Byte_to_BCD(gsRTCbuf.month));
    RTC_RegWrite(0x04, Byte_to_BCD(gsRTCbuf.day));
    RTC_RegWrite(0x02, Byte_to_BCD(gsRTCbuf.hour));
    RTC_RegWrite(0x01, Byte_to_BCD(gsRTCbuf.mins));

    RTC_RegWrite(0x00, 0x80);        // Start the clock (ST = 1), seconds = 00.

    return 0;
}

/*
 * Synchronize system RTC time & date with hardware RTCC.
 * If the operation is successful, return 0;  else return error code (negative int).
 */
int  RTCC_Synchronize( void )
{
    // First, check if RTCC is responding...
    if (RTC_RegWrite(0x20, 0xA5) == ERROR) return (-1);  // Write SRAM reg
    if (RTC_RegRead(0x20) != 0xA5) return (-2);          // Read back & verify

    gsRTCbuf.secs = BCD_to_binary(RTC_RegRead(0x00) & 0x7F);
    gsRTCbuf.mins = BCD_to_binary(RTC_RegRead(0x01) & 0x7F);
    gsRTCbuf.hour = BCD_to_binary(RTC_RegRead(0x02) & 0x3F);
    gsRTCbuf.day = BCD_to_binary(RTC_RegRead(0x04)) & 0x3F;
    gsRTCbuf.month = BCD_to_binary(RTC_RegRead(0x05) & 0x1F);
    gsRTCbuf.year = BCD_to_binary(RTC_RegRead(0x06));

    RTC_date_time_set();
    return 0;
}

#endif // RTCC_TYPE_MCP79410


/*
 * Write RTC register -- address 0xDE on I2C2 bus
 * Entry args: regaddr = RTC register address (0x00..0x13)
 *             regdata = register data to write
 * Returns:    error code (0: no error, -1: I2C bus error)
 */
int  RTC_RegWrite(uint8 regaddr, uint8 regdata)
{
    loop = 0;
    if (!I2C2MasterStart(0xDE)) return (-1);
    loop = 1;
    if (!I2C2MasterSend(regaddr)) return (-1);
    loop = 2;
    if (!I2C2MasterSendAckStop(regdata)) return (-1);
    loop = 3;

    return 0;
}

/*
 * Read RTC register -- ISL1208 (U9) -- address 0xDE on I2C2 bus
 * Entry args: regaddr = RTC register address (0x00..0x13)
 *
 * Returns:    regdata (byte value 0x00..0xFF) if read success, or...
 *             error code (-1) if I2C bus error occurred
 */
int  RTC_RegRead(uint8 regaddr)
{
    uint8 regdata;

    if (!I2C2MasterStart(0xDE)) return (-1);
    if (!I2C2MasterSend(regaddr)) return (-1);
    if (!I2C2MasterReadNackStop(0xDE, &regdata, 1)) return (-1);

    return regdata;
}


// Increment a 2-digit modulo-60 packed BCD number (byte) and return the result
uint8  BCDincrMod60(uint8 bcd_byte)
{
    int digit0, digit1;
    uint8 carry = 0;

    digit0 = (bcd_byte & 0xF) + 1;
    if (digit0 > 9) { digit0 = 0; carry = 1; }
    digit1 = (bcd_byte >> 4) + carry;
    if (digit1 > 5) digit1 = 0;  // rollover to 00

    return (digit1 << 4) + digit0;
}

// Convert binary number (byte) to a 2-digit modulo-60 packed BCD number (returned)
uint8  Byte_to_BCD(uint8 bin_byte)
{
    uint8 digit0, digit1;

    bin_byte = bin_byte % 60;
    digit0 = bin_byte % 10;
    digit1 = bin_byte / 10;

    return (digit1 << 4) + digit0;
}

// Convert 2-digit BCD number (byte) to binary number (returned)
uint8  BCD_to_binary(uint8 BCD_byte)
{
    return (BCD_byte >> 4) * 10 + (BCD_byte & 0xF);
}

//=================================================================================================

#ifdef  USE_CONSOLE_CLI
/**
*   CLI command function:  Cmnd_rtcc
*
*   The "rtcc" command provides diagnostics for a hardware RTCC device.
*/
void  Cmnd_rtcc( int argCount, char *argValue[] )
{
    int    errorCode = 0;
    char   option = tolower(argValue[1][1]);
    uint8  seconds, regData;

    if ( argCount == 1 || *argValue[1] == '?' )   // help wanted
    {
        putstr( "RTCC device diagnostics... \n");
        putstr( "Usage:  rtcc <opt> \n");
        putstr( "Options ----------------------------- \n");
        putstr( "   -i  :  Initialize RTCC device \n");
        putstr( "   -t  :  Check Timer is running \n" );
        putstr( "   -s  :  Show Seconds reg. \n");
        putstr( "   -r  :  Read control & status reg's \n");
        putstr( "Use 'time' cmd to read/set system RTC.\n");
        return;
    }

    if (RTC_RegWrite(SCRATCH_REG_ADDR, 0xA5) == ERROR)  // Write SRAM reg
    {
        putstr("! RTCC not responding (I2C bus error?)\n");
        return;
    }
    if (RTC_RegRead(SCRATCH_REG_ADDR) != 0xA5)  // Read back & verify
    {
        putstr("! RTCC register R/W error.\n");
        return;
    }

    switch (option)
    {
        case 'i':  // Home
        {
            errorCode = RTCC_Initialize();
            if (errorCode == -1) putstr("! RTCC not responding (I2C bus error?)\n"); else
            if (errorCode == -2) putstr("! RTCC register R/W error.\n"); else
            if (errorCode == -3) putstr("! RTCF bit not reset (ISL1208)\n");
            break;
        }
        case 't':
        {
            seconds = RTC_RegRead(0x00) & 0x7F;
            WaitMilliseconds(2000);
            ////
            if ((RTC_RegRead(0x00) & 0x7F) == seconds)  // not counting
                putstr("! RTCC not running.\n");
            else  putstr("* RTCC running.\n");
            break;
        }
        case 's':
        {
            seconds = BCD_to_binary(RTC_RegRead(0x00) & 0x7F);
            putch('\t');  putDecimal(seconds, 2);
            NEW_LINE;
            break;
        }
        case 'r':
        {
            regData = RTC_RegRead(CONTROL_REG_ADDR);
            putstr("Config reg = $");  putHexByte(regData);
            regData = RTC_RegRead(STATUS_REG_ADDR);
            putstr(", Status reg = $");  putHexByte(regData);
            NEW_LINE;
            break;
        }
        default:  break;

    } // end switch
}

#endif  // USE_CONSOLE_CLI

#endif // defined RTCC_TYPE_ISL1208 || defined RTCC_TYPE_MCP79410
