/*________________________________________________________________________________________*\

    File:        console.h

    Author:      M.J.Bauer

    Header file for console CLI module.
\*________________________________________________________________________________________*/

#ifndef  _CONSOLE_H_
#define  _CONSOLE_H_

#include "system_def.h"
#include "Drivers/UART_drv.h"

/******************************************************************************************\
*            APPLICATION-SPECIFIC DEFINITIONS REQUIRED BY THE CONSOLE CLI
*/
#define  CLI_MAX_ARGS       12        // Maximum # of Command Line arguments (incl. name)
#define  CMND_LINE_MAX_LEN  76        // Maximum length of command string
#define  CMD_HIST_BUF_SIZE  20        // Maximum number of commands in recall buffer
#define  MAX_COMMANDS       250       // Maximum number of CLI commands

#ifdef BUILD_REMI_CONTROLLER_APP
#define CLI_VERSION_TITLE  "REMI Command Line User Interface...\n"
#endif
#ifdef BUILD_LCD_GRAPHICS_DEMO
#define  CLI_VERSION_TITLE  "--- LCD Graphics Demo and Utility --- \n"
#endif
//
/******************************************************************************************/

#ifndef  SU_PASSWORD
#define  SU_PASSWORD      "xxxx"     // Default "super user" password
#endif

//                External functions required by the console CLI
//
#ifdef CLI_USING_UART1
#define RxDataAvail()   UART1_RxDataAvail()
#define RxFlush()       UART1_RxFlush()
#define getch()         UART1_getch()
#define putch(b)        UART1_putch(b)
#define putstr(s)       UART1_putstr(s)     
#endif

#ifdef CLI_USING_UART2
#define RxDataAvail()   UART2_RxDataAvail()
#define RxFlush()       UART2_RxFlush()
#define getch()         UART2_getch()
#define putch(b)        UART2_putch(b)
#define putstr(s)       UART2_putstr(s)     
#endif

#if !defined (CLI_USING_UART1) && !defined (CLI_USING_UART2)
#error "Can't build console-CLI module without serial port UART defined!"
#endif

uint32  millisecTimer();     // Return value of free-running 32-bit counter (unit = 1ms)
void    BackgroundTaskExec(void);  // Background task executive located in main module

#define  kbhit()     RxDataAvail()    // alias
#define  NEW_LINE    putstr("\n")

/******************************************************************************************/


#ifndef ASCII_NUL  // Common ASCII control codes

#define  ASCII_NUL       0
#define  ASCII_ACK       6
#define  ASCII_BS        8        /* Ctrl+H, Backspace */
#define  ASCII_HT        9
#define  ASCII_TAB       9
#define  ASCII_LF       10
#define  ASCII_CR       13
#define  ASCII_DC2      18        /* Ctrl+R, Device Control 2 */
#define  ASCII_NAK      21
#define  ASCII_CAN      24        /* Ctrl+X, Cancel line */
#define  ASCII_ESC      27
#define  ASCII_SP       32

#endif


enum  eDebugDataType
{
    HEX_BYTE,
    CHAR_BYTE,
    DEC_BYTE,
    HEX_WORD,
    DEC_UWORD,
    DEC_IWORD,
    HEX_LONG,
    DEC_ILONG
};

typedef  void (*CLIfunc)( int argc, char *argv[] );        // pointer to CLI function


// Command table entry looks like this
struct  CmndTableEntry_t
{
    char    *phzName;            // pointer to command name (string literal)
    uint8    yAttribute;        // flag:  1 => end-user (visible); 0 => debug (hidden)
    CLIfunc  Function;            // pointer to CLI function
};

#define  GEN_CMD     'G'
#define  APP_CMD     'A'
#define  DEBUG_CMD   'D'

typedef struct UserSettableParameters  // For "param" (alias "set") command
{
    char   name[12];     // param nick-name, max. 10 chars (last entry = "$")
    char   format;       // display format: 'r' == real, 'i' == integer
    float *address;      // pointer to global parameter
    float  minValue;     // minimum acceptable value
    float  maxValue;     // maximum acceptable value

} UserSettableParameter_t;


// Prototypes of functions defined in console CLI module...

void    ConsoleCLI_Service( void );
void    CommandLineInterpreter( void );
void    PrepareForNewCommand( void );
void    EnterCommandInHistory( void );
void    RecallCommand( void );
void    EraseLine( void );
uint8   strmatch( char *pcStr1, const char *phzStr2 );
int     getstr(char *strBuf, int maxLen);

void    OutputParamValue(char *nickName, float fValue);
void    ListParamNamesValues(void);
void    SetParameterValue(char *nickName, float fValue);
char    SuperUserAccess(void);

// ------------  Generic CLI commands  ----------------------

void    Cmnd_default( int argCount, char * argValue[] );
void    Cmnd_diag( int argCount, char * argValue[] );
void    Cmnd_flags( int argCount, char * argValue[] );
void    Cmnd_help( int argCount, char * argValue[] );
void    Cmnd_note( int argCount, char * argValue[] );
void    Cmnd_list( int argCount, char * argValue[] );
void    Cmnd_param( int argCount, char * argValue[] );
void    Cmnd_reset( int argCount, char * argValue[] );
void    Cmnd_ver( int argCount, char * argValue[] );
void    Cmnd_watch( int argCount, char * argValue[] );
void    Cmnd_su( int argCount, char * argValue[] );
void    Cmnd_time( int argCount, char * argValue[] );
void    Cmnd_rtcc( int argCount, char *argValue[] );

// --------  File-System and I/O Device CLI commands  -------

void    Cmnd_format( int argCount, char * argValue[] );
void    Cmnd_ls( int argCount, char * argValue[] );
void    Cmnd_rm( int argCount, char * argValue[] );
void    Cmnd_view( int argCount, char * argValue[] );
void    Cmnd_fcre8( int argCount, char * argValue[] );

void    Cmnd_sst25( int argCount, char * argValue[] );
void    Cmnd_sdf( int argCount, char * argValue[] );
void    Cmnd_export(int argCount, char * argVal[]);
void    Cmnd_import(int argCount, char * argVal[]);

// ----------  Application-Specific CLI commands  -----------

#ifdef BUILD_SOUND_FX_SYNTH_APP
void    Cmnd_synth( int argCount, char * argValue[] );
void    Cmnd_play( int argCount, char * argValue[] );
#endif
#ifdef BUILD_LCD_GRAPHICS_DEMO
void    Cmnd_grcfix( int argCount, char * argValue[] );
void    Cmnd_fontmod(int argCount, char * argValue[]);
#endif
#ifdef BUILD_REMI_CONTROLLER_APP
void    Cmnd_wav( int argCount, char * argValue[] );
void    Cmnd_mpr121( int argCount, char * argValue[] );
#endif

// ---------------- Debug commands  -------------------------

void    Cmnd_dump( int argCount, char * argValue[] );
void    Cmnd_peek( int argCount, char * argValue[] );
void    Cmnd_poke( int argCount, char * argValue[] );
void    Cmnd_pio( int argCount, char * argValue[] );

// Formatted numeric string output functions...

void    putBoolean( uint8 );
void    putHexDigit( uint8 );
void    putHexByte( uint8 );
void    putHexWord( uint16 );
void    putHexLong( uint32 );
void    putDecimal( int32 lVal, uint8 bSize );

// Character & string numeric conversion functions...

uint8   dectobin( char c );
uint16  decatoi( char * pac, int8 bNdigs );
uint32  long_decatoi( char * pac, int8 bNdigs );
uint8   hexctobin( char c );
uint16  hexatoi( char * s );
uint32  long_hexatoi( char * s );
uint8   isHexDigit( char c );
char    hexitoc( short wDigit );
void    wordToHexStr( uint16 wVal, char *pcResult );
void    longToDecimalStr( int32 lVal, char *pcResult, uint8 bFieldSize );

#endif    /* _CONSOLE_H_ */
